/* convert intel hex to viewlogic rom format */

#include <stdio.h>

extern char * disasm(int byte, int memory_location);



/* parses a line of intel hex code, stores the data in bytes[] */
/* and the beginning address in addr, and returns a 1 if the */
/* line was valid, or a 0 if an error occured.  The variable */
/* num gets the number of bytes that were stored into bytes[] */

int parse_hex_line(theline, bytes, addr, num, code)
char *theline;
int *addr, *num, *code, bytes[];
{
	int sum, len, cksum;
	char *ptr;
	
	*num = 0;
	if (theline[0] != ':') return 0;
	if (strlen(theline) < 11) return 0;
	ptr = theline+1;
	if (!sscanf(ptr, "%02x", &len)) return 0;
	ptr += 2;
	if ( strlen(theline) < (11 + (len * 2)) ) return 0;
	if (!sscanf(ptr, "%04x", addr)) return 0;
	ptr += 4;
	  /* printf("Line: length=%d Addr=%d\n", len, *addr); */
	if (!sscanf(ptr, "%02x", code)) return 0;
	ptr += 2;
	sum = (len & 255) + ((*addr >> 8) & 255) + (*addr & 255) + (*code & 255);
	while(*num != len) {
		if (!sscanf(ptr, "%02x", &bytes[*num])) return 0;
		ptr += 2;
		sum += bytes[*num] & 255;
		(*num)++;
		if (*num >= 256) return 0;
	}
	if (!sscanf(ptr, "%02x", &cksum)) return 0;
	if ( ((sum & 255) + (cksum & 255)) & 255 ) return 0; /* checksum error */
	return 1;
}

/* loads an intel hex file into the global memory[] array */
/* the command is a string such as "L file.hex" */

main(argc, argv)
int argc;
char *argv[];
{
	char filename[200], outfile[200];
	char line[1000];
	FILE *fin, *fout;
	int addr, n, status, bytes[256];
	int i, lineno=1;

	if (argc < 2) {
		fprintf(stderr, "Usage: hex2rom file\n");
		exit(1);
	}
	strcpy(filename, argv[1]);
	strcpy(outfile, argv[1]);
	strcat(filename, ".hex");
	strcat(outfile, ".rom");
	fin = fopen(filename, "r");
	if (fin == NULL) {
		fprintf(stderr, "Can't open file '%s' for reading.\n", filename);
		exit(1);
	}
	fout = fopen(outfile, "w");
	if (fout == NULL) {
		fprintf(stderr, "Can't open file '%s' for writing.\n", outfile);
		exit(1);
	}
	fprintf(fout, "| Viewlogic ROM data file\n");
	fprintf(fout, "| Created using hex2rom, part\n");
	fprintf(fout, "| of the OSU-8 programming tools,\n");
	fprintf(fout, "| by Paul Stoffregen, March 1995\n");
	fprintf(fout, "bitrange 7:0\n");
	fprintf(fout, "default radix hex\n");
	while (!feof(fin) && !ferror(fin)) {
		line[0] = '\0';
		fgets(line, 1000, fin);
		if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = '\0';
		if (line[strlen(line)-1] == '\r') line[strlen(line)-1] = '\0';
		if (parse_hex_line(line, bytes, &addr, &n, &status)) {
			if (status == 0) {  /* data */
				fprintf(fout, "adrs %d\n", addr);
				for(i=0; i<=(n-1); i++) {
					fprintf(fout, "%02X\n", bytes[i]);
					addr++;
				}
			}
			if (status == 1) {  /* end of file */
				fclose(fin);
				exit (0);
			}
			if (status == 2) ;  /* begin of file */
		} else {
			fprintf(stderr, "Error: '%s', line: %d\n", filename, lineno);
		}
		lineno++;
	}
	fprintf(stderr, "Error: unexpended end of file\n");
	exit(1);
}


