%{
/* OSU-8 Instructional Microprocessor Assembler */
/* by Paul Stoffregen, March 1995 */
#include <stdio.h>

/* REL4() computes branch operand from dest and memory */
/* location of the jump instruction itself */
/* this is later adjusted by one for jcu, of course */
#define BRANCH_SPACING 4
#define REL4(dest, mem) (((dest)-((((mem)+1)/(\
BRANCH_SPACING))*(BRANCH_SPACING)))/(BRANCH_SPACING))

extern void yyrestart(FILE *new_file);
extern char * disasm(int byte, int memory_location);
extern void hexout(int byte, int memory_location, int end);

%}

%token ADD AND CALL CLR CPL DEC INC
%token JCD JCU JUMP LOAD MOVE NOP OR POP PUSH
%token RET RETI RL RLC RR RRC SET STORE SUB

%token JC JNC JV JNV JZ JNZ JP JNP DJNZ CJNE CJE

%token A B AT AB BT BB SP P2 P2L P2H
%token P1 P1L P1H P1LB P1LT P1HB P1HT
%token IN0 IN1 OUT0 OUT1 OUT2 OUT3
%token C V Z P

%token WORD NUMBER HEXNUM1 HEXNUM2 BINNUM CHAR SPCHAR STRING

%token ORG EQU DB

%%

all:           line
             | line all;

line:          linesymbol ':' linenosym {
			if (p1) build_sym_list(symbol_name);
			if (p2) assign_value(symbol_name, mem);
			mem += m_len; }
             | linenosym {mem += m_len;}

linenosym:     directive
             | instruction  {m_len=1; if (p3) out(opcode);}
             | macro

directive:     ORG expression {mem = expr_result; m_len = 0;}
             | EQU equsymbol ',' expression { 
			if (p1) build_sym_list(symbol_name);
			if (p2 && expr_ok)
				assign_value(symbol_name, expr_result);
			m_len = 0; }
             | DB bytes


bytes:

linesymbol:    WORD  { 
			strcpy(symbol_name, yytext);
			if (is_target(symbol_name)) pad_with_nop();
		}

equsymbol:     WORD  {
			strcpy(symbol_name, yytext);
		}

relative:      WORD {
			if (p1) build_target_list(yytext);
			if (p3 && !is_def(yytext)) {
				fprintf(stderr, "Error: branch to undefined");
				fprintf(stderr, " symbol '%s'", yytext);
				fprintf(stderr, " in line %d\n", lineno);
				exit(1);
			}
			if (p3) {
				jump_dest = get_value(yytext);
				operand = REL4(jump_dest, mem);
				if (operand < 0) {
					operand *= -1;
					operand -= 1;  /* jcu 0 = jump -4 up */
				}
				 /* will be offset, regardless of up or down */
			}
			if (p3 && operand > 15) boob_error();
		}

macro_relative: WORD {  /* special for macros, since other inst before jump */
			/* so we just load jump_dest now and use it later on */
			if (p1) build_target_list(yytext);
			if (p3 && !is_def(yytext)) {
				fprintf(stderr, "Error: branch to undefined");
				fprintf(stderr, " symbol '%s'", yytext);
				fprintf(stderr, " in line %d\n", lineno);
				exit(1);
			}
			if (p3) jump_dest = get_value(yytext);
		}

data:          '#' expression

expression:    NUMBER {
			sscanf(yytext, "%d", &expr_result);
			expr_ok = 1; }
             | HEXNUM1 {
			sscanf(yytext, "%*c%*c%x", &expr_result);
			expr_ok = 1; }
             | HEXNUM2 {
			sscanf(yytext, "%x%*[hH]", &expr_result);
			expr_ok = 1; }
             | BINNUM {  /* sscanf doesn't do binary... */
			expr_result = binary2int(yytext);
			expr_ok = 1; }
             | CHAR   {
			expr_result = yytext[1];
			expr_ok = 1; }
             | SPCHAR {
			switch (yytext[2]) {
				case 'n':
				case 'N': expr_result = 10; break;
				case 'r':
				case 'R': expr_result = 13; break;
				case '0': expr_result = 0; break;
				default:
					fprintf(stderr, "Unknown special ");
					fprintf(stderr, "%s in line %d\n", \
					  yytext, lineno);
					exit(1);
			}
			expr_ok = 1;
		}
             | WORD {
			expr_ok = is_def(yytext);
			if (expr_ok) expr_result = get_value(yytext);
		}

macro:         MOVE data ',' A  {
			m_len=2;
			if (p3) {
				out(0x80 + (expr_result & 15));
				mem++;
				out(0x90 + ((expr_result & 240) >> 4));
				mem--;
			}
		}

             | MOVE data ',' B  {
			m_len=2;
			if (p3) {
				out(0xA0 + (expr_result & 15));
				mem++;
				out(0xB0 + ((expr_result & 240) >> 4));
				mem--;
			}
		}

             | MOVE data ',' P1 {
			m_len=4;
			if (p3) {
				out(0xC0 + (expr_result & 15));
				mem++;
				out(0xD0 + ((expr_result & 240) >> 4));
				mem++;
				out(0xE0 + ((expr_result & 3840) >> 8));
				mem++;
				out(0xF0 + ((expr_result & 61440) >> 12));
				mem -= 3;
			}
		}
             | MOVE data ',' P2 {
			m_len=9;
			if (p3) {
				out(0x46); mem++;  /* PUSH  P1H */
				out(0x44); mem++;  /* PUSH  P1L */
				out(0xC0 + (expr_result & 15)); mem++;
				out(0xD0 + ((expr_result & 240) >> 4)); mem++;
				out(0xE0 + ((expr_result & 3840) >> 8)); mem++;
				out(0xF0 + ((expr_result & 61440) >> 12)); mem++;
				out(0x75); mem++;  /* MOVE  P1 -> P2 */
				out(0x45); mem++;  /* POP   P1L */
				out(0x47);         /* POP   P1H */
				mem -= 8;
			}
		}
             | JC macro_relative {
			m_len=1;
			if (p3) do_jump_on_carry();
		}
             | JNC macro_relative {
			m_len=2;
			if (p3) {
				out(0x67);  /* CPL C */
				mem++;
				do_jump_on_carry();
				mem--;
			}
		}
             | JV macro_relative {
			m_len=2;
			if (p3) {
				out(0x62);  /* MOV N->C */
				mem++;
				do_jump_on_carry();
				mem--;
			}
		}
             | JNV macro_relative {
			m_len=3;
			if (p3) {
				out(0x62);  /* MOV N->C */
				mem++;
				out(0x67);  /* CPL C */
				mem++;
				do_jump_on_carry();
				mem -= 2;
			}
		}
             | JZ macro_relative {
			m_len=2;
			if (p3) {
				out(0x63);  /* MOV Z->C */
				mem++;
				do_jump_on_carry();
				mem--;
			}
		}
             | JNZ macro_relative {
			m_len=3;
			if (p3) {
				out(0x63);  /* MOV Z->C */
				mem++;
				out(0x67);  /* CPL C */
				mem++;
				do_jump_on_carry();
				mem -= 2;
			}
		}
             | JP macro_relative {
			m_len=2;
			if (p3) {
				out(0x64);  /* MOV P->C */
				mem++;
				do_jump_on_carry();
				mem--;
			}
		}
             | JNP macro_relative {
			m_len=3;
			if (p3) {
				out(0x64);  /* MOV P->C */
				mem++;
				out(0x67);  /* CPL C */
				mem++;
				do_jump_on_carry();
				mem -= 2;
			}
		}
             | CJNE A ',' B ',' macro_relative {
			m_len=7;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0x65); mem++;  /* CLR   C */
				out(0x15); mem++;  /* SUB   A-B -> B */
				out(0x43); mem++;  /* POP   B */
				out(0x63); mem++;  /* MOVE  Z -> C */
				out(0x67); mem++;  /* CPL   C */
				do_jump_on_carry();
				mem -= 6;
			}
		}
             | CJE A ',' B ',' macro_relative {
			m_len=6;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0x65); mem++;  /* CLR   C */
				out(0x15); mem++;  /* SUB   A-B -> B */
				out(0x43); mem++;  /* POP   B */
				out(0x63); mem++;  /* MOVE  Z -> C */
				do_jump_on_carry();
				mem -= 5;
			}
		}
             | CJNE A ',' data ',' macro_relative {
			m_len=9;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x65); mem++;  /* CLR   C */
				out(0x15); mem++;  /* SUB   A-B -> B */
				out(0x43); mem++;  /* POP   B */
				out(0x63); mem++;  /* MOVE  Z -> C */
				out(0x67); mem++;  /* CPL   C */
				do_jump_on_carry();
				mem -= 8;
			}
		}
             | CJE A ',' data ',' macro_relative {
			m_len=8;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x65); mem++;  /* CLR   C */
				out(0x15); mem++;  /* SUB   A-B -> B */
				out(0x43); mem++;  /* POP   B */
				out(0x63); mem++;  /* MOVE  Z -> C */
				do_jump_on_carry();
				mem -= 7;
			}
		}
             | CJNE B ',' data ',' macro_relative {
			m_len=9;
			if (p3) {
				out(0x40); mem++;  /* PUSH  A */
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x65); mem++;  /* CLR   C */
				out(0x05); mem++;  /* SUB   A-B -> A */
				out(0x41); mem++;  /* POP   A */
				out(0x63); mem++;  /* MOVE  Z -> C */
				out(0x67); mem++;  /* CPL   C */
				do_jump_on_carry();
				mem -= 8;
			}
		}
             | CJE B ',' data ',' macro_relative {
			m_len=8;
			if (p3) {
				out(0x40); mem++;  /* PUSH  A */
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x65); mem++;  /* CLR   C */
				out(0x05); mem++;  /* SUB   A-B -> A */
				out(0x41); mem++;  /* POP   A */
				out(0x63); mem++;  /* MOVE  Z -> C */
				out(0x67); mem++;  /* CPL   C */
				do_jump_on_carry();
				mem -= 7;
			}
		}
             | DJNZ A ',' macro_relative {
			m_len=4;
			if (p3) {
				out(0x0D); mem++;  /* DEC   A */
				out(0x63); mem++;  /* MOVE  Z -> C */
				out(0x67); mem++;  /* CPL   C */
				do_jump_on_carry();
				mem -= 3;
			}
		}
             | DJNZ B ',' macro_relative {
			m_len=4;
			if (p3) {
				out(0x1F); mem++;  /* DEC   B */
				out(0x63); mem++;  /* MOVE  Z -> C */
				out(0x67); mem++;  /* CPL   C */
				do_jump_on_carry();
				mem -= 3;
			}
		}
             | ADD A '+' data ',' A {
			m_len=5;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x04); mem++;  /* ADD   A+B -> A */
				out(0x43);         /* POP   B */
				mem -= 4;
			}
		}
             | SUB A '-' data ',' A {
			m_len=5;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x05); mem++;  /* SUB   A+B -> A */
				out(0x43);         /* POP   B */
				mem -= 4;
			}
		}
             | AND A '&' data ',' A {
			m_len=5;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x06); mem++;  /* AND   A+B -> A */
				out(0x43);         /* POP   B */
				mem -= 4;
			}
		}
             | OR A '|' data ',' A {
			m_len=5;
			if (p3) {
				out(0x42); mem++;  /* PUSH  B */
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x07); mem++;  /* OR    A+B -> A */
				out(0x43);         /* POP   B */
				mem -= 4;
			}
		}
             | ADD B '+' data ',' B {
			m_len=5;
			if (p3) {
				out(0x40); mem++;  /* PUSH  A */
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x14); mem++;  /* ADD   A+B -> B */
				out(0x41);         /* POP   A */
				mem -= 4;
			}
		}
             | SUB B '-' data ',' B {
			m_len=5;
			if (p3) {
				out(0x40); mem++;  /* PUSH  A */
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x16); mem++;  /* SUB   A+B -> B */
				out(0x41);         /* POP   A */
				mem -= 4;
			}
		}
             | AND B '&' data ',' B {
			m_len=5;
			if (p3) {
				out(0x40); mem++;  /* PUSH  A */
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x16); mem++;  /* AND   A+B -> B */
				out(0x41);         /* POP   A */
				mem -= 4;
			}
		}
             | OR B '|' data ',' B {
			m_len=5;
			if (p3) {
				out(0x40); mem++;  /* PUSH  A */
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x17); mem++;  /* OR    A+B -> B */
				out(0x41);         /* POP   A */
				mem -= 4;
			}
		}
             | ADD A '+' data ',' B {
			m_len=3;
			if (p3) {
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x14);  /* ADD A+B -> B */
				mem -= 2;
			}
		}
             | SUB A '-' data ',' B {
			m_len=3;
			if (p3) {
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x15);  /* SUB A+B -> B */
				mem -= 2;
			}
		}
             | AND A '&' data ',' B {
			m_len=3;
			if (p3) {
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x16);  /* AND A+B -> B */
				mem -= 2;
			}
		}
             | OR A '|' data ',' B {
			m_len=3;
			if (p3) {
				out(0xA0 + (expr_result & 15)); mem++;
				out(0xB0 + ((expr_result >> 4) & 15)); mem++;
				out(0x17);  /* OR A+B -> B */
				mem -= 2;
			}
		}
             | ADD B '+' data ',' A {
			m_len=3;
			if (p3) {
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x04);  /* ADD A+B -> A */
				mem -= 2;
			}
		}
             | SUB B '-' data ',' A {
			m_len=3;
			if (p3) {
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x05);  /* SUB A+B -> A */
				mem -= 2;
			}
		}
             | AND B '&' data ',' A {
			m_len=3;
			if (p3) {
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x06);  /* AND A+B -> A */
				mem -= 2;
			}
		}
             | OR B '|' data ',' A {
			m_len=3;
			if (p3) {
				out(0x80 + (expr_result & 15)); mem++;
				out(0x90 + ((expr_result >> 4) & 15)); mem++;
				out(0x07);  /* OR A+B -> A */
				mem -= 2;
			}
		}









instruction:   MOVE A ',' A		{opcode = 0x00;}
             | MOVE A ',' B		{opcode = 0x10;}
             | CPL A			{opcode = 0x01;}
             | CPL A ',' B		{opcode = 0x11;}
             | MOVE B ',' A		{opcode = 0x02;}
             | MOVE B ',' B		{opcode = 0x12;}
             | CPL B ',' A		{opcode = 0x03;}
             | CPL B			{opcode = 0x13;}
             | ADD A '+' B ',' A	{opcode = 0x04;}
             | ADD A '+' B ',' B	{opcode = 0x14;}
             | SUB A '-' B ',' A	{opcode = 0x05;}
             | SUB A '-' B ',' B	{opcode = 0x15;}
             | AND A '&' B ',' A	{opcode = 0x06;}
             | AND A '&' B ',' B	{opcode = 0x16;}
             | OR A '|' B ',' A		{opcode = 0x07;}
             | OR A '|' B ',' B		{opcode = 0x17;}
             | RR A			{opcode = 0x08;}
             | RR A ',' B		{opcode = 0x18;}
             | RL A			{opcode = 0x09;}
             | RL A ',' B		{opcode = 0x19;}
             | RRC A			{opcode = 0x0A;}
             | RRC A ',' B		{opcode = 0x1A;}
             | RLC A			{opcode = 0x0B;}
             | RLC A ',' B		{opcode = 0x1B;}
             | INC A			{opcode = 0x0C;}
             | INC A ',' B		{opcode = 0x1C;}
             | DEC A			{opcode = 0x0D;}
             | DEC A ',' B		{opcode = 0x1D;}
             | INC B ',' A		{opcode = 0x0E;}
             | INC B			{opcode = 0x1E;}
             | DEC B ',' A		{opcode = 0x0F;}
             | DEC B			{opcode = 0x1F;}

             | PUSH A			{opcode = 0x40;}
             | PUSH B			{opcode = 0x42;}
             | PUSH P1L			{opcode = 0x44;}
             | PUSH P1H			{opcode = 0x46;}
             | PUSH P2L			{opcode = 0x48;}
             | PUSH P2H			{opcode = 0x4A;}
             | POP A			{opcode = 0x41;}
             | POP B			{opcode = 0x43;}
             | POP P1L			{opcode = 0x45;}
             | POP P1H			{opcode = 0x47;}
             | POP P2L			{opcode = 0x49;}
             | POP P2H			{opcode = 0x4B;}
             | CALL '@' P1		{opcode = 0x4C;}
             | JUMP '@' P1		{opcode = 0x4D;}
             | RET			{opcode = 0x4E;}
             | RETI			{opcode = 0x4F;}

             | LOAD '@' P1 ',' A	{opcode = 0x50;}
             | LOAD '@' P2 ',' A	{opcode = 0x51;}
             | LOAD '@' P1 ',' B	{opcode = 0x52;}
             | LOAD '@' P2 ',' B	{opcode = 0x53;}
             | STORE A ',' '@' P1	{opcode = 0x58;}
             | STORE A ',' '@' P2	{opcode = 0x59;}
             | STORE B ',' '@' P1	{opcode = 0x5A;}
             | STORE B ',' '@' P2	{opcode = 0x5B;}

             | LOAD P1 ',' A            {opcode = 0x50;}
             | LOAD P2 ',' A            {opcode = 0x51;}
             | LOAD P1 ',' B            {opcode = 0x52;}
             | LOAD P2 ',' B            {opcode = 0x53;}
             | STORE A ',' P1           {opcode = 0x58;}
             | STORE A ',' P2           {opcode = 0x59;}
             | STORE B ',' P1           {opcode = 0x5A;}
             | STORE B ',' P2           {opcode = 0x5B;}


             | MOVE IN0 ',' C		{opcode = 0x60;}
             | MOVE IN1 ',' C		{opcode = 0x61;}
             | MOVE V ',' C		{opcode = 0x62;}
             | MOVE Z ',' C		{opcode = 0x63;}
             | MOVE P ',' C		{opcode = 0x64;}
             | CLR C			{opcode = 0x65;}
             | SET C			{opcode = 0x66;}
             | CPL C			{opcode = 0x67;}
             | MOVE C ',' OUT0		{opcode = 0x68;}
             | MOVE C ',' OUT1		{opcode = 0x69;}
             | MOVE C ',' OUT2		{opcode = 0x6A;}
             | MOVE C ',' OUT3		{opcode = 0x6B;}
             | NOP			{opcode = 0x6E;}

             | INC P1			{opcode = 0x70;}
             | INC P2			{opcode = 0x71;}
             | DEC P1			{opcode = 0x72;}
             | DEC P2			{opcode = 0x73;}
             | MOVE P2 ',' P1		{opcode = 0x74;}
             | MOVE P1 ',' P2		{opcode = 0x75;}
             | MOVE SP ',' P1		{opcode = 0x76;}
             | MOVE P1 ',' SP		{opcode = 0x77;}
             | MOVE P1L ',' A		{opcode = 0x78;}
             | MOVE P1H ',' A		{opcode = 0x79;}
             | MOVE P1L ',' B		{opcode = 0x7A;}
             | MOVE P1H ',' B		{opcode = 0x7B;}
             | MOVE A ',' P1L		{opcode = 0x7C;}
             | MOVE A ',' P1H		{opcode = 0x7D;}
             | MOVE B ',' P1L		{opcode = 0x7E;}
             | MOVE B ',' P1H		{opcode = 0x7F;}

             | JCU relative		{opcode = 0x20 + operand;}
             | JCD relative		{opcode = 0x30 + operand;}

             | MOVE data ',' AB		{opcode = 0x80 + (expr_result & 15);}
             | MOVE data ',' AT		{opcode = 0x90 + (expr_result & 15);}
             | MOVE data ',' BB		{opcode = 0xA0 + (expr_result & 15);}
             | MOVE data ',' BT		{opcode = 0xB0 + (expr_result & 15);}
             | MOVE data ',' P1LB	{opcode = 0xC0 + (expr_result & 15);}
             | MOVE data ',' P1LT	{opcode = 0xD0 + (expr_result & 15);}
             | MOVE data ',' P1HB	{opcode = 0xE0 + (expr_result & 15);}
             | MOVE data ',' P1HT	{opcode = 0xF0 + (expr_result & 15);}
%%

/* a linked list of all the symbols */

struct symbol {
	char *name;
	int value;
	int istarget;   /* 1 if a branch target, 0 otherwise */
	int isdef;	/* 1 if defined, 0 if no value yet */
	int line_def;   /* line in which is was defined */
	struct symbol *next; };

/* a list of all the symbols that are branch targets */
/* (and will need to get aligned on 4 byte boundries) */

struct target {
	char *name;
	struct target *next; };


/* yucky global variables... but I don't care */

FILE *flst, *fhex, *fmem;
extern FILE *yyin;
extern char *yytext;
struct symbol *sym_list=NULL;
struct target *targ_list=NULL;
int lineno=1, mem=0;   /* mem is location in memory */
int p1=0, p2=0, p3=0;
int expr_result, expr_ok, m_len, jump_dest;
int opcode, operand;
char symbol_name[1000];


/* add symbols to list when we find their definition in pass #1 */
/* we will evaluate their values in pass #2, and figure out if */
/* they are branch targets betweem passes 1 and 2.  Every symbol */
/* should appear exactly once in this list, since it can't be redefined */

struct symbol * build_sym_list(thename)
char *thename;
{
	struct symbol *new, *p;

/*	printf("  Symbol: %s  Line: %d\n", thename, lineno); */
	new = (struct symbol *) malloc(sizeof(struct symbol));
	new->name = (char *) malloc(strlen(thename)+1);
	strcpy(new->name, thename);
	new->value = 0;
	new->istarget = 0;
	new->isdef = 0;
	new->line_def = lineno;
	new->next = NULL;
	if (sym_list == NULL) return (sym_list = new);
	p = sym_list;
	while (p->next != NULL) p = p->next;
	p->next = new;
	return (new);
}

int assign_value(thename, thevalue)
char *thename;
int thevalue;
{
	struct symbol *p;

	p = sym_list;
 	while (p != NULL) {
		if (!(strcasecmp(thename, p->name))) {
			p->value = thevalue;
			p->isdef = 1;
			return (0);
		}
		p = p->next;
	}
	fprintf(stderr, "Internal Error!  Couldn't find symbol\n");
	exit(1);
}

int get_value(thename)
char *thename;
{
	struct symbol *p;
	p = sym_list;
	while (p != NULL) {
		if (!(strcasecmp(thename, p->name)))
			return (p->value);
		p = p->next;
	}
	fprintf(stderr, "Internal Error!  Couldn't find symbol value\n");
	exit(1);
}
		


/* add every branch target to this list as we find them */
/* ok if multiple entries of same symbol name in this list */

struct target * build_target_list(thename)
char *thename;
{
	struct target *new, *p;
	new = (struct target *) malloc(sizeof(struct target));
	new->name = (char *) malloc(strlen(thename)+1);
	strcpy(new->name, thename);
	new->next = NULL;
	if (targ_list == NULL) return (targ_list = new);
	p = targ_list;
	while (p->next != NULL) p = p->next;
	p->next = new;
	return (new);
}

/* figure out which symbols are branch targets */

void flag_targets()
{
	struct symbol *p_sym;
	struct target *p_targ;
	p_targ = targ_list;
	while (p_targ != NULL) {
		p_sym = sym_list;
		while (p_sym != NULL) {
			if (!strcasecmp(p_sym->name, p_targ->name))
				p_sym->istarget = 1;
			p_sym = p_sym->next;
		}
		p_targ = p_targ->next;
	}
}

print_symbol_table()
{
	struct symbol *p;
	p = sym_list;
	while (p != NULL) {
		printf("Symbol: %12s  Value: %5d (%04X)  Def: ", \
		  p->name, p->value, p->value);
		if (p->isdef) printf("Yes"); else printf("No ");
		printf("  Target: ");
		if (p->istarget) printf("Yes"); else printf("No ");
		printf("  Line %d\n", p->line_def);
		p = p->next;
	}
}

/* check that every symbol is in the table only once */

void check_redefine()
{
	struct symbol *p1, *p2;
	p1 = sym_list;
	while (p1 != NULL) {
		p2 = p1->next;
		while (p2 != NULL) {
			if (!strcasecmp(p1->name, p2->name)) {
				fprintf(stderr, "Error: symbol '%s' redefined on line %d", p1->name, p2->line_def);
				fprintf(stderr, ", first defined on line %d\n", p1->line_def);
			exit(1);
			}
			p2 = p2->next;
		}
		p1 = p1->next;
	}
}

int is_target(thename)
char *thename;
{
	struct symbol *p;
	p = sym_list;
	while (p != NULL) {
		if (!strcasecmp(thename, p->name)) return (p->istarget);
		p = p->next;
	}
	return (0);
}

int is_def(thename)
char *thename;
{
	struct symbol *p;
	p = sym_list;
	while (p != NULL) {
		if (!strcasecmp(thename, p->name) && p->isdef) return(1);
		p = p->next;
	}
	return (0);
}

void out(opcode)
int opcode;
{
	printf("%04X: %02X   %s\n", mem, opcode, disasm(opcode, mem));
	hexout(opcode, mem, 0);
}


/* add NOPs to align memory location on a valid branch target address */

void pad_with_nop()
{
	while (mem % BRANCH_SPACING) {
		if (p3) out(opcode = 0x6E);
		mem++;
	}
}

/* print branch out of bounds error */

void boob_error()
{
	fprintf(stderr, "Error: branch out of bounds");
	fprintf(stderr, " in line %d\n", lineno);
	exit(1);
}

/* output the jump either direction on carry */
/* jump_dest and mem must have the proper values */

void do_jump_on_carry()
{
	if (p3) {
		operand = REL4(jump_dest, mem);
		if (operand < 0) {
			operand *= -1;
			operand -= 1;
			if (operand > 15) boob_error;
			out(0x20 + (operand & 15));
		} else {
			if (operand > 15) boob_error;
			out(0x30 + (operand & 15));
		}
	}
}


/* turn a string like "10010110b" into an int */

int binary2int(str)
char *str;
{
	register int i, j=1, sum=0;
	
	for (i=strlen(str)-2; i >= 0; i--) {
		sum += j * (str[i] == '1');
		j *= 2;
	}
	return (sum);
}

void print_usage();

/* pass #1 (p1=1) find all symbol defs and branch target names */
/* pass #2 (p2=1) align branch targets, evaluate all symbols */
/* pass #3 (p3=1) produce object code */

main(argc, argv)
int argc;
char *argv[];
{
	char infilename[200], outfilename[200], pstr;

	if (argc < 2) print_usage();
	strcpy(infilename, argv[1]);
	if(strlen(infilename) > 4) {
		if (strncasecmp(infilename+strlen(infilename)-4, ".asm", 4))
			strcat(infilename, ".asm");
	} else strcat(infilename, ".asm");
	strcpy(outfilename, infilename);
	outfilename[strlen(outfilename)-4] = '\0';
	strcat(outfilename, ".hex");
	yyin = fopen(infilename, "r");
	if (yyin == NULL) {
		fprintf(stderr, "Can't open file '%s'.\n", infilename);
		exit(1);
	}
	fhex = fopen(outfilename, "w");
	if (fhex == NULL) {
		fprintf(stderr, "Can't write file '%s'.\n", outfilename);
		exit(1);
	}
	printf("\n\nOSU-8 Instructional Microprocessor Assembler\n");
	printf("Paul Stoffregen, March 1995.\n\n");
	printf("Building Symbol Table:\n");
	p1 = 1;
	mem = 0;
	yyparse();
	flag_targets();
	/* print_symbol_table(); */
	check_redefine();
	p1 = 0;
	p2 = 1;
	rewind(yyin);
	yyrestart(yyin);
	lineno = 1;
	printf("Aligning Branch Targets:\n");
	mem = 0;
	yyparse();
	print_symbol_table();
	p2 = 0;
	p3 = 1;
	rewind(yyin);
	yyrestart(yyin);
	lineno = 1;
	printf("Generating Object Code:\n");
	mem = 0;
	yyparse();
	fclose(yyin);
	hexout(0, 0, 1);  /* flush and close intel hex file output */
	exit(0);

}



yyerror(s)
char *s;
{
	fprintf(stderr, "%s near '%s', line %d\n", s, yytext, lineno);
	exit(1);
}

void print_usage()
{
	fprintf(stderr, "Usage: osu8asm file\n");
	fprintf(stderr, "   or  osu8asm file.asm\n");
	exit(1);
}

int yywrap()
{
	return 1;
}
