/* disassembler routine used in both osu8asm and osu8sim */
/* the idea behind having a separate disassembler when */
/* either program could easily just output the correct */
/* instruction text is so that bugs will (hopefully) be */
/* much easier to notice... a bug in the assember which */
/* makes the wrong opcode can't somehow print the right */
/* instruction text next to it.  */

/* also in here is the code to make an intel hex file */

#include <stdio.h>

extern FILE *fhex;  /* the file to put intel hex into */

char * disasm(byte, memory_location)
int byte, memory_location;
{

static int opcode;
static char buffer[100];
static char *instruction[]={
	"MOVE  A -> A",
	"CPL   A",
	"MOVE  B -> A",
	"CPL   B -> A",
	"ADD   A+B -> A",
	"SUB   A-B -> A",
	"AND   A&B -> A",
	"OR    A|B -> A",
	"RR    A",
	"RL    A",
	"RRC   A",
	"RLC   A",
	"INC   A",
	"DEC   A",
	"INC   B -> A",
	"DEC   B -> A",
	"MOVE  A -> B",
	"CPL   A -> B",
	"MOVE  B -> B",
	"CPL   B",
	"ADD   A+B -> B",
	"SUB   A-B -> B",
	"AND   A&B -> B",
	"OR    A|B -> B",
	"RR    A -> B",
	"RL    A -> B",
	"RRC   A -> B",
	"RLC   A -> B",
	"INC   A -> B",
	"DEC   A -> B",
	"INC   B",
	"DEC   B",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"J",
	"PUSH  A",
	"POP   A",
	"PUSH  B",
	"POP   B",
	"PUSH  P1L",
	"POP   P1L",
	"PUSH  P1H",
	"POP   P1H",
	"PUSH  P2L",
	"POP   P2L",
	"PUSH  P2H",
	"POP   P2H",
	"CALL  @P1",
	"JUMP  @P1",
	"RET",
	"RETI",
	"LOAD  @P1 -> A",
	"LOAD  @P2 -> A",
	"LOAD  @P1 -> B",
	"LOAD  @P2 -> B",
	"<undefined>",
	"<undefined>",
	"<undefined>",
	"<undefined>",
	"STORE A -> @P1",
	"STORE A -> @P2",
	"STORE B -> @P1",
	"STORE B -> @P2",
	"<undefined>",
	"<undefined>",
	"<undefined>",
	"<undefined>",
	"MOVE  IN0 -> C",
	"MOVE  IN1 -> C",
	"MOVE  V -> C",
	"MOVE  Z -> C",
	"MOVE  P -> C",
	"CLR   C",
	"SET   C",
	"CPL   C",
	"MOVE  C -> OUT0",
	"MOVE  C -> OUT1",
	"MOVE  C -> OUT2",
	"MOVE  C -> OUT3",
	"<undefined>",
	"<undefined>",
	"NOP",
	"MOV   C -> IE",
	"INC   P1",
	"INC   P2",
	"DEC   P1",
	"DEC   P2",
	"MOVE  P2 -> P1",
	"MOVE  P1 -> P2",
	"MOVE  SP -> P1",
	"MOVE  P1 -> SP",
	"MOVE  P1L -> A",
	"MOVE  P1H -> A",
	"MOVE  P1L -> B",
	"MOVE  P1H -> B",
	"MOVE  A -> P1L",
	"MOVE  A -> P1H",
	"MOVE  B -> P1L",
	"MOVE  B -> P1H"};

	opcode = (byte >> 4) & 15;
	if (opcode == 0 || opcode == 1 || opcode == 4 \
	   || opcode == 5 || opcode == 6 || opcode == 7)
		return (instruction[byte & 127]);
	if (opcode == 2) {
		sprintf(buffer, "JCU   %04X", \
		 ((((memory_location + 1) / 4) * 4) - ((byte & 15) + 1) * 4) & 65535);
		return (buffer);
	}
	if (opcode == 3) {
		sprintf(buffer, "JCD   %04X", \
		 ((((memory_location + 1) / 4) * 4) + (byte & 15) * 4) & 65535);
		return (buffer);
	}
	sprintf(buffer, "MOVE  #%01X -> ", (byte & 15));
	switch (opcode) {
		case 8: strcat(buffer, "Ab"); break;
		case 9: strcat(buffer, "At"); break;
		case 10: strcat(buffer, "Bb"); break;
		case 11: strcat(buffer, "Bt"); break;
		case 12: strcat(buffer, "P1Lb"); break;
		case 13: strcat(buffer, "P1Lt"); break;
		case 14: strcat(buffer, "P1Hb"); break;
		case 15: strcat(buffer, "P1Ht"); break;
	}
	return (buffer);
}

/* produce intel hex file output */

#define MAXHEXLINE 32	/* the maximum number of bytes to put in one line */

void hexout(byte, memory_location, end)
int byte, memory_location, end;
{
	static int byte_buffer[MAXHEXLINE];
	static int last_mem, buffer_pos, buffer_addr;
	static int writing_in_progress=0;
	register int i, sum;

	if (!writing_in_progress) {
		/* initial condition setup */
		last_mem = memory_location-1;
		buffer_pos = 0;
		buffer_addr = memory_location;
		writing_in_progress = 1;
		}

	if ( (memory_location != (last_mem+1)) || (buffer_pos >= MAXHEXLINE) \
	 || ((end) && (buffer_pos > 0)) ) {
		/* it's time to dump the buffer to a line in the file */
		fprintf(fhex, ":%02X%04X00", buffer_pos, buffer_addr);
		sum = buffer_pos + ((buffer_addr>>8)&255) + (buffer_addr&255);
		for (i=0; i < buffer_pos; i++) {
			fprintf(fhex, "%02X", byte_buffer[i]&255);
			sum += byte_buffer[i]&255;
		}
		fprintf(fhex, "%02X\n", (-sum)&255);
		buffer_addr = memory_location;
		buffer_pos = 0;
	}

	if (end) {
		fprintf(fhex, ":00000001FF\n");  /* end of file marker */
		fclose(fhex);
		writing_in_progress = 0;
	}
		
	last_mem = memory_location;
	byte_buffer[buffer_pos] = byte & 255;
	buffer_pos++;
}


