%{
#include "ucode.tab.h"
#include <stdio.h>
#include "ucode.h"
extern int lineno;
extern int input_defined, output_defined, state_defined, action_defined;
extern int definition_defined, doing_playback;
int ret_token;

extern struct input_list *current_input;
extern struct output_list *current_output;
extern struct state_list *current_state;
extern struct action_list *current_action;
extern struct definition_list *current_definition;
extern struct token_list *current_token;

static int rec_mode=0;

struct input_list *query_input;
struct output_list *query_output;
struct state_list *query_state;
struct action_list *query_action;
struct definition_list *query_definition;

#include "ucodell.h"
#include "ucodedebug.h"

void begin_rec_defn();
void end_rec_defn();
void rec();
void begin_playback();
int playback();		/* this actually is a fake yylex for macro substitution */
int my_yylex();

%}


%%


input_section			return INPUT_TAG;
output_section			return OUTPUT_TAG;
state_section			return STATE_TAG;
action_section			return ACTION_TAG;
definition_section		return DEFINITION_TAG;
ucode_section			return UCODE_TAG;

default				return DEFAULT;
if				{if (rec_mode) rec(IF, yytext);
				 return IF;}
code				{if (rec_mode) rec(CODE, yytext);
				 return CODE;}
bits				{if (rec_mode) rec(BITS, yytext);
				 return BITS;}
begin_def			return BEGIN_DEF;
end_def				return END_DEF;
:				{if (rec_mode) rec(COLON, yytext);
				 return COLON;}
=[ \t]*[xX?]			{if (rec_mode) rec(UNKNOWN, yytext);
				 return UNKNOWN;}
=				{if (rec_mode) rec(EQUAL, yytext);
				 return EQUAL;}
\;				{if (rec_mode) rec(SEMICOLON, yytext);
				 return SEMICOLON;}
,				{if (rec_mode) rec(COMMA, yytext);
				 return COMMA;}
\$				{if (rec_mode) rec(DOLLAR, yytext);
				 return DOLLAR;}




[_~a-zA-Z][-a-zA-Z0-9_+>~]*   {	if (input_defined) {
					query_input=find_input(yytext);
					if (query_input != NULL) {
						current_input = query_input;
						if (rec_mode) rec(INPUT, yytext);
						return INPUT;
					}
				}
				if (output_defined) {
					query_output=find_output(yytext);
					if (query_output != NULL) {
						current_output = query_output;
						if (rec_mode) rec(OUTPUT, yytext);
						return OUTPUT;
					}
				}
                                if (state_defined) {
                                        query_state=find_state(yytext);
                                        if (query_state != NULL) {
						current_state = query_state;
						if (rec_mode) rec(STATE, yytext);
                                                return STATE;
					}
                                }
                                if (action_defined) {
                                        query_action=find_action(yytext);
                                        if (query_action != NULL) {
						current_action = query_action;
						if (rec_mode) rec(ACTION, yytext);
                                                return ACTION;
					}
                                }
				if (definition_defined) {
					query_definition=find_definition(yytext);
					if (query_definition != NULL) {
						current_definition=query_definition;
						current_token = \
							current_definition->tokens;
						if (current_token != NULL) {
							yytext=current_token->text;
							ret_token=current_token->token;
							playback_globals(ret_token);
							current_token=current_token->next;
							doing_playback = \
								(current_token != NULL);
							return (ret_token);
						}
					}
				}
				if (rec_mode) rec(WORD, yytext);
				return WORD; }


[01]+[bB]			{if (rec_mode) rec(BINNUM, yytext);
				 return BINNUM; }
0[xX][0-9a-fA-F]+		{if (rec_mode) rec(HEXNUM1, yytext);
				 return HEXNUM1; }
[0-9a-fA-F]+[hH]		{if (rec_mode) rec(HEXNUM2, yytext);
				 return HEXNUM2; }
-?[0-9]+			{if (rec_mode) rec(NUMBER, yytext);
				 return NUMBER; }
\"[^"\n]*["\n]			{if (rec_mode) rec(STRING, yytext);
				 return STRING; }


\#.*\n		++lineno;
[ \t]		;
\n		++lineno;
.		return yytext[0];

%%

void begin_rec_defn(the_definition)
struct definition_list *the_definition;
{
	rec_mode = 1;
	current_token = the_definition->tokens;
	if (current_token != NULL) {
		while (current_token->next != NULL) {
			current_token = current_token->next;
		}
	}
}

void end_rec_defn()
{
	rec_mode = 0;
	current_token = NULL;
}

void rec(the_token, the_text)
int the_token;
char *the_text;
{
	struct token_list *p, *new;

	new = (struct token_list *) malloc(sizeof(struct token_list));
	new->token = the_token;
	new->text = (char *) malloc(strlen(the_text)+1);
	strcpy(new->text, the_text);
	new->value = 0;
	new->next = NULL;
	if (current_token == NULL) current_definition->tokens = new;
	else current_token->next = new;
	current_token = new;

}

int playback()
{
	int token;

	if (current_token == NULL) {
		fprintf(stderr, "Fatal Error in definition substitution.");
		exit (1);
	}
	token = current_token->token;
	yytext = current_token->text;
	playback_globals(token);
	current_token = current_token->next;
	if (current_token == NULL) {
		doing_playback = 0;
	}
	return (token);
}

int my_yylex()
{
	if (doing_playback) return( playback() );
	else return( yylex() );
}


playback_globals(the_token)
int the_token;
{
	if (the_token == INPUT) current_input = find_input(yytext);
	if (the_token == OUTPUT) current_output = find_output(yytext);
	if (the_token == STATE) current_state = find_state(yytext);
	if (the_token == ACTION) current_action = find_action(yytext);
}

