%{
/* microcode interpreter: convert ucode to truth table */


#define YYSTACKSIZE 8000  /* will overflow using default settting */
			  /* with large defines, like inst decoding */

#include <stdio.h>
#include "ucode.h"      /* all the struct definitions */
#include "ucodell.h"
#include "ucodedebug.h"
#include "ucodeglobals.h"

int input_defined=0, output_defined=0, state_defined=0, action_defined=0;
int definition_defined=0, doing_playback=0;
int current_expression, cond_value;

struct input_list *current_input;
struct output_list *current_output;
struct state_list *current_state, *uc_cur_state;
struct action_list *current_action;
struct definition_list *current_definition;
struct token_list *current_token;


extern void begin_rec_defn();
extern void end_rec_defn();


%}

%token INPUT_TAG OUTPUT_TAG STATE_TAG
%token ACTION_TAG DEFINITION_TAG UCODE_TAG
%token DEFAULT EQUAL IF BITS CODE UNKNOWN BEGIN_DEF END_DEF
%token COLON COMMA SEMICOLON DOLLAR
%token WORD NUMBER HEXNUM1 HEXNUM2 BINNUM STRING
%token INPUT OUTPUT STATE ACTION


%%

all:		input_section {input_defined=1;}
		output_section {output_defined=1;}
		state_section {state_assign(); state_defined=1;}
		action_section {action_defined=1;}
		definition_section {definition_defined=1;}
		ucode_section 


input_section:	INPUT_TAG inputs | INPUT_TAG
inputs:		input
		| input inputs
input:		input_name SEMICOLON
		| input_name input_args SEMICOLON
input_name:	WORD {current_input=add_input(yytext); }
input_args:	input_arg 
		| input_arg input_args
input_arg:	BITS EQUAL expression {current_input->numbits=$3;}


output_section:	OUTPUT_TAG outputs | OUTPUT_TAG
outputs:	output | output outputs
output:		output_name SEMICOLON
		| output_name out_defs SEMICOLON
output_name:	WORD {current_output=add_output(yytext);}
out_defs:	out_def | out_def out_defs
out_def:	DEFAULT EQUAL expression {current_output->def=$3;}
		| DEFAULT UNKNOWN {current_output->def=-1;}
		| BITS EQUAL expression {current_output->numbits=$3;}


state_section:	STATE_TAG states | STATE_TAG
states:		state | state states
state:		state_name SEMICOLON
		| state_name state_args SEMICOLON
state_name:	WORD {current_state=add_state(yytext);}
state_args:	state_arg | state_arg state_args
state_arg:	CODE EQUAL expression {current_state->code=$3;}


action_section:	| ACTION_TAG actions | ACTION_TAG
actions:	action | action actions
action:		action_name SEMICOLON
		| action_name act_outs SEMICOLON
action_name:	WORD {current_action=add_action(yytext);}
act_outs:	act_out | act_out act_out_sep act_outs
act_out:	OUTPUT EQUAL expression {
			add_action_out(current_action, current_output->name, $3);}
		| OUTPUT UNKNOWN {
			add_action_out(current_action, current_output->name, -1);}
act_out_sep:	| COMMA


definition_section:
		| DEFINITION_TAG definitions | DEFINITION_TAG
definitions:	definition | definition definitions
definition:	defn_name BEGIN_DEF {begin_rec_defn(current_definition);}
		defn {end_rec_defn();} END_DEF
defn_name:	WORD {current_definition=add_definition(yytext);}
defn:		defn_rule
		| defn_rule COMMA
		| defn_rule rule_sep defn
defn_rule:	ACTION
		| OUTPUT EQUAL expression
		| OUTPUT UNKNOWN
		| condition ACTION
		| condition OUTPUT EQUAL expression
		| condition OUTPUT UNKNOWN



ucode_section:	UCODE_TAG uc_states | UCODE_TAG
uc_states:	uc_state | uc_state uc_states
uc_state:	uc_state_name COLON rules SEMICOLON
		| uc_state_name COLON SEMICOLON
uc_state_name:	STATE {uc_cur_state=current_state;}
rules:		rule
		| rule COMMA
		| rule rule_sep rules
rule: 		ACTION { add_uncond_action(uc_cur_state, current_action); }
		| OUTPUT EQUAL expression { add_uncond_out(uc_cur_state, \
			current_output, current_expression); }
		| OUTPUT UNKNOWN { add_uncond_out(uc_cur_state, \
			current_output, -1); }
		| condition ACTION {add_cond_action(uc_cur_state, \
			current_action, current_input, cond_value);}
		| condition OUTPUT EQUAL expression 
			{add_cond_out(uc_cur_state, current_output, \
			current_expression, current_input, cond_value);}
		| condition OUTPUT UNKNOWN
			{add_cond_out(uc_cur_state, current_output, \
			-1, current_input, cond_value);}
rule_sep:	| COMMA
condition:	IF '(' INPUT EQUAL cond_exp ')'
		| IF INPUT EQUAL cond_exp
		| IF '(' INPUT EQUAL EQUAL cond_exp ')'
		| IF INPUT EQUAL EQUAL cond_exp
cond_exp:	expression {cond_value = $1;}


expression:	NUMBER { sscanf(yytext, "%d", &$$);
			current_expression=$$;}
       		| HEXNUM1 { sscanf(yytext, "%*c%*c%x", &$$);
			current_expression=$$;}
       		| HEXNUM2 { sscanf(yytext, "%x%*[hH]", &$$);
			current_expression=$$;}
		| BINNUM { $$ = binary2int(yytext);
			current_expression=$$;}
		| DOLLAR STATE { $$ = current_state->code; 
			current_expression=$$;}
		/* this next line lets you omit the '$', but I */
		/* think it may cause shift/reduce conflicts if */
		/* more rules are added elsewhere.. works for now */
		| STATE { $$ = current_state->code; 
			current_expression=$$;}
	
%%


/* turn a string like "10010110b" into an int */

int binary2int(str)
char *str;
{
	register int i, j=1, sum=0;
	
	for (i=strlen(str)-2; i >= 0; i--) {
		sum += j * (str[i] == '1');
		j *= 2;
	}
	return (sum);
}

/* assign codes to any states that the user didn't */

void state_assign()
{
	struct state_list *p;
	register struct state_list *q;
	register int newcode=0, codeused;

	p = begin_state_list;
	while (p != NULL) {
		if (p->code == -1) {
			codeused=1;
			while (codeused) {
				codeused=0;
				q = begin_state_list;
				while (q != NULL) {
					if (q->code == newcode) {
						newcode++;
						codeused = 1;
						break;
					}
					q = q->next;
				}
			}
		p->code = newcode++;
		}
	p = p->next;
	}
}


