#include <stdio.h>
#include "ucodell.h"
#include "ucodedebug.h"
#include "ucode.h"
#include "ucodeglobals.h"

void blif_output();
void blif_print_inputs();
void blif_print_outputs();
void blif_cleanup_tt_string(char *s);



void blif_output()
{
	int i, out_num=0;
        struct output_list *p_output;
	struct truth_table *p_tt;
        char name[] = "control";
	char *c;

	/* print the header information */
	fprintf(fout, ".model %s\n", name);
	fprintf(fout, ".inputs ");
	blif_print_inputs();
	fprintf(fout, "\n");
	fprintf(fout, ".outputs ");
        blif_print_outputs();
        fprintf(fout, "\n");


	/* clean up the truth table strings to make things easier */
        p_tt = begin_truth_table;
        while (p_tt != NULL) {
                blif_cleanup_tt_string(p_tt->input);
		for (c = p_tt->input; *c != '\0'; c++)
			if (*c == 'X') *c = '-';
                blif_cleanup_tt_string(p_tt->output);
                p_tt = p_tt->next;
        }

	/* print pla table logic */
	p_output = begin_output_list;
	out_num = 0;
	while (p_output != NULL) {
		/* print 1's pla table */
		if (p_output->numbits == 1) {	/* single output bit */
			fprintf(fout, ".names ");
			blif_print_inputs();
			fprintf(fout, " %s\n", p_output->name);
			p_tt = begin_truth_table;
			while (p_tt != NULL) {
				if (p_tt->output[out_num] == '1')
					fprintf(fout, "%s 1\n", p_tt->input);
				p_tt = p_tt->next;
			}
			out_num++;
		} else {			/* multi-bit output */
                        for (i=p_output->numbits-1; i>=0; i--) {
				fprintf(fout, ".names ");
				blif_print_inputs();
				fprintf(fout, " %s%d\n", p_output->name, i);
				p_tt = begin_truth_table;
				while (p_tt != NULL) {
					if (p_tt->output[out_num] == '1')
						fprintf(fout, "%s 1\n", p_tt->input);
					p_tt = p_tt->next;
				}
				out_num++;
			}
		}
		p_output = p_output->next;
	}

	/* print don't case cases */
	fprintf(fout, ".exdc\n");
	p_output = begin_output_list;
	out_num = 0;
	while (p_output != NULL) {
                if (p_output->numbits == 1) {   /* single output bit */
                        fprintf(fout, ".names ");
                        blif_print_inputs();
                        fprintf(fout, " %s\n", p_output->name);
                        p_tt = begin_truth_table;
                        while (p_tt != NULL) {
                                if (p_tt->output[out_num] == 'X')
                                        fprintf(fout, "%s 1\n", p_tt->input);
                                p_tt = p_tt->next;
                        }
                        out_num++;
                } else {                        /* multi-bit output */
                        for (i=p_output->numbits-1; i>=0; i--) {
                                fprintf(fout, ".names ");
                                blif_print_inputs();
                                fprintf(fout, " %s%d\n", p_output->name, i);
                                p_tt = begin_truth_table;
                                while (p_tt != NULL) {
                                        if (p_tt->output[out_num] == 'X')
                                                fprintf(fout, "%s 1\n", p_tt->input);
                                        p_tt = p_tt->next;
                                }
				out_num++;
                        }
                }
                p_output = p_output->next;
        }

	fprintf(fout, ".end\n");
	fflush(fout);
}


void blif_print_inputs()
{
	struct input_list *p_input;
	register int i;

	for (i=state_numbits-1; i>=0; i--) fprintf(fout, "state%d ", i);

        p_input = begin_input_list;
        while (p_input != NULL) {
                if (p_input->numbits == 1) {
                        fprintf(fout, "%s", p_input->name);
                } else {
			for (i=p_input->numbits-1; i>=0; i--) {
                        	fprintf(fout, "%s%d", p_input->name, i);
				if (i != 0) fprintf(fout, " ");
			}
                }
                p_input = p_input->next;
		if (p_input != NULL) fprintf(fout, " ");
        }
}


void blif_print_outputs()
{
	struct output_list *p_output;
	register int i;

        p_output = begin_output_list;
        while (p_output != NULL) {
                if (p_output->numbits == 1) {
                        fprintf(fout, "%s", p_output->name);
                } else {
			for (i=p_output->numbits-1; i>=0; i--) {
                        	fprintf(fout, "%s%d", p_output->name, i);
				if (i != 0) fprintf(fout, " ");
			}
                }
                p_output = p_output->next;
                if (p_output != NULL) fprintf(fout, " ");
        }
}



void blif_cleanup_tt_string(char *s)
{
        register char *p1, *p2;

        p1 = p2 = s;
        while (*p2 != '\0') {
                if (*p2 == '0' || *p2 == '1' || *p2 == 'X') {
                        if (p1 != p2) *p1 = *p2;
                        p1++;
                }
                p2++;
        }
        *p1 = '\0';
}

