#include <stdio.h>
#include "ucode.h"
#include "ucodeglobals.h"

struct input_list * add_input(thename)
char *thename;
{
	struct input_list *new;
	register struct input_list *p;

	new = (struct input_list *) malloc(sizeof(struct input_list));
	new->name = (char *) malloc(strlen(thename)+1);
	strcpy(new->name, thename);
	new->numbits = 1;  /* default one bit only */
	new->next = NULL;
	if (begin_input_list == NULL) return (begin_input_list = new);
	p = begin_input_list;
	while (p->next != NULL) p = p->next;
	p->next = new;
	return (new);
}

struct input_list * find_input(thename)
char *thename;
{
	register struct input_list *p;

	p = begin_input_list;
 	while (p != NULL) {
		if (!(strcasecmp(thename, p->name))) return (p);
		p = p->next;
	}
	return (NULL);
}

struct output_list * add_output(thename)
char *thename;
{
	struct output_list *new;
	register struct output_list *p;

	new = (struct output_list *) malloc(sizeof(struct output_list));
	new->name = (char *) malloc(strlen(thename)+1);
	strcpy(new->name, thename);
	new->numbits = 1;	/* default one bit only */
	new->def = -1;		/* default is unknown */
	new->current_value = -1;
	new->next = NULL;
	if (begin_output_list == NULL) return (begin_output_list = new);
	p = begin_output_list;
	while (p->next != NULL) p = p->next;
	p->next = new;
	return (new);
}

struct output_list * find_output(thename)
char *thename;
{
        register struct output_list *p;

        p = begin_output_list;
        while (p != NULL) {
                if (!(strcasecmp(thename, p->name))) return (p);
                p = p->next;
        }
        return (NULL);
}


struct state_list * add_state(thename)
char *thename;
{
        struct state_list *new;
	register struct state_list *p;

        new = (struct state_list *) malloc(sizeof(struct state_list));
        new->name = (char *) malloc(strlen(thename)+1);
        strcpy(new->name, thename);
	new->code = -1;
	new->jobs = NULL;
	new->def_outs = NULL;
        new->next = NULL;
        if (begin_state_list == NULL) return (begin_state_list = new);
        p = begin_state_list;
        while (p->next != NULL) p = p->next;
        p->next = new;
        return (new);
}

struct state_list * find_state(thename)
char *thename;
{
        register struct state_list *p;

        p = begin_state_list;
        while (p != NULL) {
                if (!(strcasecmp(thename, p->name))) return (p);
                p = p->next;
        }
        return (NULL);
}



struct action_list * add_action(thename)
char *thename;
{
        struct action_list *new;
	register struct action_list *p;

        new = (struct action_list *) malloc(sizeof(struct action_list));
        new->name = (char *) malloc(strlen(thename)+1);
        strcpy(new->name, thename);
	new->outputs = NULL;
        new->next = NULL;
        if (begin_action_list == NULL) return (begin_action_list = new);
        p = begin_action_list;
        while (p->next != NULL) p = p->next;
        p->next = new;
        return (new);
}

struct action_list * find_action(thename)
char *thename;
{
	register struct action_list *p;

        p = begin_action_list;
        while (p != NULL) {
                if (!(strcasecmp(thename, p->name))) return (p);
                p = p->next;
        }
        return (NULL);
}


struct definition_list * add_definition(thename)
char *thename;
{
        struct definition_list *new;
	register struct definition_list *p;

        new = (struct definition_list *) malloc(sizeof(struct definition_list));
        new->name = (char *) malloc(strlen(thename)+1);
        strcpy(new->name, thename);
        new->tokens = NULL;
        new->next = NULL;
        if (begin_definition_list == NULL) return (begin_definition_list = new);
        p = begin_definition_list;
        while (p->next != NULL) p = p->next;
        p->next = new;
        return (new);
}

struct definition_list * find_definition(thename)
char *thename;
{
	register struct definition_list *p;

        p = begin_definition_list;
        while (p != NULL) {
                if (!(strcasecmp(thename, p->name))) return (p);
                p = p->next;
        }
        return (NULL);
}

/* this one assumes no other code will add to the truth table entries, so that it */
/* may keep a static pointer to the end of the list.  This avoids searching the */
/* list for every addition we make, which is pretty desirable since the list is */
/* going to be pretty long, even for a simple microcode */

struct truth_table * add_truth_table(the_input, the_output)
char *the_input;
char *the_output;
{
	struct truth_table *new;
	static struct truth_table *p_end=NULL;  /* assume this is the only routine to add to the list */

	new = (struct truth_table *) malloc(sizeof(struct truth_table));
	new->input = (char *) malloc(strlen(the_input)+1);
	strcpy(new->input, the_input);
	new->output = (char *) malloc(strlen(the_output)+1);
	strcpy(new->output, the_output);
	new->next = NULL;
	if (begin_truth_table == NULL) return (begin_truth_table = p_end = new);
	return (p_end = p_end->next = new);
}


struct action_outs * add_action_out(the_action, the_output_name, the_value)
struct action_list *the_action;
char *the_output_name;
int the_value;
{
	struct action_outs *new;
	register struct action_outs *p;

	new = (struct action_outs *) malloc(sizeof(struct action_outs));
	new->output = find_output(the_output_name);
	if (new->output == NULL) undef_output_err(the_output_name);
	new->value = the_value;
	if (the_action->outputs == NULL) return(the_action->outputs = new);
	p = the_action->outputs;
	while (p->next != NULL) p = p->next;
	p->next = new;
	return (new);
}



void add_uncond_out(the_state, the_output, the_value)
struct state_list *the_state;
struct output_list *the_output;
int the_value;
{
	struct action_outs *new;
	register struct action_outs *p;

	new = (struct action_outs *) malloc(sizeof(struct action_outs));
	new->output = the_output;
	new->value = the_value;
	new->next = NULL;
	if (the_state->def_outs == NULL) the_state->def_outs = new;
	else {
		p = the_state->def_outs;
		while (p->next != NULL) p = p->next;
		p->next = new;
	}
}

void add_uncond_action(the_state, the_action)
struct state_list *the_state;
struct action_list *the_action;
{
	register struct action_outs *p;

	p = the_action->outputs;
	while (p != NULL) {
		add_uncond_out(the_state, p->output, p->value);
		p = p->next;
	}
}

void add_cond_out(the_state, the_output, the_value, the_input, the_cond)
struct state_list *the_state;
struct output_list *the_output;
struct input_list *the_input;
int the_value, the_cond;
{
	struct job_list *jnew;
	register struct job_list *jptr;
	struct action_outs *new;
	register struct action_outs *p;

	new = (struct action_outs *) malloc(sizeof(struct action_outs));
	new->output = the_output;
	new->value = the_value;
	new->next = NULL;

	jptr = the_state->jobs;
	while (jptr != NULL) {
		if ((jptr->cond_input == the_input) && \
		  (jptr->cond_value == the_cond)) break;
		jptr = jptr->next;
	}
	if (jptr != NULL) {
		/* in this case, the condition already existed in */
		/* our list, so we just add this output spec to its list */
		p=jptr->tasks;
		if (p == NULL) p = new;
		else {
			while (p->next != NULL)	p = p->next;
			p->next = new;
		}
	} else {
		/* in this case, we add the condition to the condition */
		/* list and start its output list with this one */
		jnew = (struct job_list *) malloc(sizeof(struct job_list));
		jnew->cond_input = the_input;
		jnew->cond_value = the_cond;
		jnew->tasks = new;
		jnew->next = NULL;
		if (the_state->jobs == NULL) the_state->jobs = jnew;
		else {
			jptr = the_state->jobs;
			while (jptr->next != NULL) jptr = jptr->next;
			jptr->next = jnew;
		}
	}
}
		
void add_cond_action(the_state, the_action, the_input, the_cond)
struct state_list *the_state;
struct action_list *the_action;
struct input_list *the_input;
int the_cond;
{
        register struct action_outs *p;

        p = the_action->outputs;
        while (p != NULL) {
                add_cond_out(the_state, p->output, p->value, \
		  the_input, the_cond);
                p = p->next;
        }
}




