
#include <stdio.h>
#include "ucodell.h"
#include "ucodedebug.h"

#include "ucode.h"

FILE *fin, *fout;
extern FILE *yyin;
extern char *yytext;
struct input_list *begin_input_list=NULL;
struct output_list *begin_output_list=NULL;
struct state_list *begin_state_list=NULL;
struct action_list *begin_action_list=NULL;
struct definition_list *begin_definition_list=NULL;
struct truth_table *begin_truth_table=NULL;
int lineno=1;



main(argc, argv)
int argc;
char *argv[];
{

	if (argc > 1) {
		fin = fopen(argv[1], "r");
		if (fin == NULL) {
			fprintf(stderr, "Can't open file '%s'.\n", argv[1]);
			exit(1);
		}
	}
	yyin = fin;

	if (argc > 2) {
		fout = fopen(argv[2], "w");
		if (fout == NULL) {
			fprintf(stderr, "Can't write file '%s'.\n", argv[2]);
			exit(1);
		}
	} else fout = stdout;

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "\nMicrocode to Truth Table, Paul Stoffregen, 1995\n");
		fprintf(stderr, " Usage:\n");
		fprintf(stderr, "     %s filename.ucode\n", argv[0]);
		fprintf(stderr, " or  %s filename.ucode filename.tt\n\n", argv[0]);
		exit(1);
	}

	if (fout != stdout) printf("Reading Input uCode\n");
	yyparse();
	fclose(yyin);

	/* print_definitions(); */
	/* print_ucode(); */

	if (fout != stdout) printf("Creating Truth Table...\n");
	create_tt();

#ifdef DO_VHDL
	if (fout != stdout) printf("Writing VHDL Output\n");
	vhdl_output();
#endif

#ifdef DO_BLIF
        if (fout != stdout) printf("Writing BLIF Output\n");
        blif_output();
#endif

        fflush(fout);
        exit(0);

}




